#ifndef _CCONFIDENCE_H_
#define _CCONFIDENCE_H_

#include "types.h"
#include "MathUtils.h"
#include <iostream>

class CConfidence
{
public:
  CConfidence(const float c = min_conf_default, const float min_c = min_conf_default,
	      const float max_c = max_conf_default);
  CConfidence(const CConfidence&);

  // Should plus and minus be defined on CConfidence? Probably!
  // Does it really "mean" anything? Yes, something...
  // What should be done with the limits?
  // Now I take the highest Max and lowest Min.
  // Can it be negative? Only if min_conf is negative
  // since all values should be between min_conf and max_conf.
  CConfidence operator -()                   const;
  CConfidence operator +(const CConfidence&) const;
  CConfidence operator +(const float)        const;
  CConfidence operator -(const CConfidence&) const;
  CConfidence operator -(const float)        const;
  CConfidence operator *(const CConfidence&) const;
  CConfidence operator *(const float)        const;
  CConfidence operator /(const CConfidence&) const;
  CConfidence operator /(const float)        const;
  
  CConfidence& operator +=(const CConfidence&);
  CConfidence& operator +=(const float);
  CConfidence& operator -=(const CConfidence&);
  CConfidence& operator -=(const float);
  CConfidence& operator *=(const CConfidence&);
  CConfidence& operator *=(const float);
  CConfidence& operator /=(const CConfidence&);
  CConfidence& operator /=(const float);

  bool operator ==(const CConfidence&) const;
  bool operator ==(const float)        const;
  bool operator !=(const CConfidence&) const;
  bool operator !=(const float)        const;
  bool operator > (const CConfidence&) const;
  bool operator > (const float)        const;
  bool operator >=(const CConfidence&) const;
  bool operator >=(const float)        const;
  bool operator < (const CConfidence&) const;
  bool operator < (const float)        const;
  bool operator <=(const CConfidence&) const;
  bool operator <=(const float)        const;
  bool operator ! ()                   const;

  CConfidence& SetMinConfidence();
  CConfidence& SetMaxConfidence();
  CConfidence& SetConfidence(const float);
  float GetConfidence() const;

  CConfidence& SetMaxConfidence(const float);
  float GetMaxConfidence() const;
  CConfidence& SetMinConfidence(const float);
  float GetMinConfidence() const;

  static float GetMaxConfDefault();
  static void  SetMaxConfDefault(const float);
  static float GetMinConfDefault();
  static void  SetMinConfDefault(const float);
  
  friend std::ostream& operator<<(std::ostream&, const CConfidence&);
  friend std::istream& operator>>(std::istream&, CConfidence&);
  
protected:
  // Class variables that keep track of the default
  // min and max confidence values.
  static float max_conf_default;
  static float min_conf_default;

  float min_conf;
  float max_conf;
  float confidence;
  
};

/*
#if CHECK_DIVISION_BY_ZERO == 1
// This constant has the default limits. Don't know if that's the correct definition.
// Since one don't want to loose the current limits, this constant
// will not be used in the code.
static const CConfidence CCONFIDENCE_DIVISION_BY_ZERO_RESULT(FLOAT_TYPE_DIVISION_BY_ZERO_RESULT);
#endif CHECK_DIVISION_BY_ZERO
*/

inline CConfidence CConfidence::operator -() const
{
  return CConfidence(-confidence, min_conf, max_conf);
} 

inline CConfidence CConfidence::operator +(const CConfidence& c) const
{
  return CConfidence(confidence + c.confidence, Min(min_conf, c.min_conf), Max(max_conf, c.max_conf));
} 

inline CConfidence CConfidence::operator +(const float d) const
{
  return CConfidence(confidence + d, min_conf, max_conf);
} 

inline CConfidence CConfidence::operator -(const CConfidence& c) const
{
  return CConfidence(confidence - c.confidence, Min(min_conf, c.min_conf), Max(max_conf, c.max_conf));
} 

inline CConfidence CConfidence::operator -(const float d) const
{
  return CConfidence(confidence - d, min_conf, max_conf);
} 

inline CConfidence CConfidence::operator *(const CConfidence& c) const
{
  return CConfidence(confidence * c.confidence, Min(min_conf, c.min_conf), Max(max_conf, c.max_conf));
} 

inline CConfidence CConfidence::operator *(const float d) const
{
  return CConfidence(confidence * d, min_conf, max_conf);
} 

inline CConfidence CConfidence::operator /(const CConfidence& c) const
{
#if CHECK_DIVISION_BY_ZERO == 1
  if (c.confidence == 0)
    return CConfidence(FLOAT_DIVISION_BY_ZERO_RESULT, Min(min_conf, c.min_conf), Max(max_conf, c.max_conf));
  else
#endif
  return CConfidence(confidence / c.confidence, Min(min_conf, c.min_conf), Max(max_conf, c.max_conf));
} 

inline CConfidence CConfidence::operator /(const float d) const
{
#if CHECK_DIVISION_BY_ZERO == 1
  if (d == 0)
    return CConfidence(FLOAT_DIVISION_BY_ZERO_RESULT, min_conf, max_conf);
  else
#endif
    return CConfidence(confidence / d, min_conf, max_conf);
} 

inline CConfidence& CConfidence::operator +=(const CConfidence& c)
{
  confidence += c.confidence;
  min_conf = Min(min_conf, c.min_conf);
  max_conf = Max(min_conf, Max(max_conf, c.max_conf));
  confidence = MinMax(min_conf, confidence, max_conf);
  return *this;
}

inline CConfidence& CConfidence::operator +=(const float d)
{
  confidence += d;
  confidence = MinMax(min_conf, confidence, max_conf);  
  return *this;
}

inline CConfidence& CConfidence::operator -=(const CConfidence& c)
{
  confidence -= c.confidence;
  min_conf = Min(min_conf, c.min_conf);
  max_conf = Max(min_conf, Max(max_conf, c.max_conf));
  confidence = MinMax(min_conf, confidence, max_conf);
  return *this;
}

inline CConfidence& CConfidence::operator -=(const float d)
{
  confidence -= d;
  confidence = MinMax(min_conf, confidence, max_conf);  
  return *this;
}

inline CConfidence& CConfidence::operator *=(const CConfidence& c)
{
  confidence *= c.confidence;
  min_conf = Min(min_conf, c.min_conf);
  max_conf = Max(min_conf, Max(max_conf, c.max_conf));
  confidence = MinMax(min_conf, confidence, max_conf);
  return *this;
}

inline CConfidence& CConfidence::operator *=(const float d)
{
  confidence *= d;
  confidence = MinMax(min_conf, confidence, max_conf);  
  return *this;
}

inline CConfidence& CConfidence::operator /=(const CConfidence& c)
{
#if CHECK_DIVISION_BY_ZERO == 1
  if (c.confidence == 0)
    confidence = FLOAT_DIVISION_BY_ZERO_RESULT;
  else
#endif
    confidence /= c.confidence;
  
  min_conf = Min(min_conf, c.min_conf);
  max_conf = Max(min_conf, Max(max_conf, c.max_conf));
  confidence = MinMax(min_conf, confidence, max_conf);
  return *this;
}

inline CConfidence& CConfidence::operator /=(const float d)
{
#if CHECK_DIVISION_BY_ZERO == 1
  if (d == 0)
    confidence = FLOAT_DIVISION_BY_ZERO_RESULT;
  else
#endif
    confidence /= d;
  
  confidence = MinMax(min_conf, confidence, max_conf);  
  return *this;
}

inline bool CConfidence::operator ==(const CConfidence& c) const
{
  return (confidence == c.confidence);
}

inline bool CConfidence::operator ==(const float d) const
{
  return (confidence == d);
}

inline bool CConfidence::operator !=(const CConfidence& c) const
{
  return (confidence != c.confidence);
}

inline bool CConfidence::operator !=(const float d) const
{
  return (confidence != d);
}

inline bool CConfidence::operator > (const CConfidence& c) const
{
  return (confidence > c.confidence);
}

inline bool CConfidence::operator > (const float d) const
{
  return (confidence > d);
}

inline bool CConfidence::operator >=(const CConfidence& c) const
{
  return (confidence >= c.confidence);
}

inline bool CConfidence::operator >=(const float d) const
{
  return (confidence >= d);
}

inline bool CConfidence::operator < (const CConfidence& c) const
{
  return (confidence < c.confidence);
}

inline bool CConfidence::operator < (const float d) const
{
  return (confidence < d);
}

inline bool CConfidence::operator <=(const CConfidence& c) const
{
  return (confidence <= c.confidence);
}

inline bool CConfidence::operator <=(const float d) const
{
  return (confidence <= d);
}

inline bool CConfidence::operator ! () const
{
  return (confidence != 0);
}

inline CConfidence& CConfidence::SetMinConfidence()
{
  confidence = min_conf;
  return *this;
}

inline CConfidence& CConfidence::SetMaxConfidence()
{
  confidence = max_conf;
  return *this;
}

inline CConfidence& CConfidence::SetConfidence(const float c)
{
  confidence = MinMax(min_conf, c, max_conf);
  return *this;
}

inline float CConfidence::GetConfidence() const
{
  return confidence;
}

inline CConfidence& CConfidence::SetMaxConfidence(const float max_c)
{
  max_conf = Max(max_c, min_conf);
  confidence = Min(confidence, max_conf);
  return *this;
}

inline float CConfidence::GetMaxConfidence() const
{
  return max_conf;
}


inline CConfidence& CConfidence::SetMinConfidence(const float min_c)
{
  min_conf = Min(min_c, max_conf);
  confidence = Max(confidence, min_conf);
  return *this;
}

inline float CConfidence::GetMinConfidence() const
{
  return min_conf;
}

inline float CConfidence::GetMaxConfDefault() 
{
  return max_conf_default;
}

inline void CConfidence::SetMaxConfDefault(const float max_c_def)
{
  max_conf_default = Max(min_conf_default, max_c_def);
}

inline float CConfidence::GetMinConfDefault()
{
  return min_conf_default;
}

inline void CConfidence::SetMinConfDefault(const float min_c_def)
{
  min_conf_default = Min(max_conf_default, min_c_def);
}
  
inline std::ostream& operator<<(std::ostream& os, const CConfidence& c)
{
  return os << c.confidence;
}

inline std::istream& operator>>(std::istream& is, CConfidence& c)
{
  is >> c.confidence;
  return is;
}

#endif
