#include "CDashCommand.h"
#include "CKnowledge.h"
#include "utils.h"

CDashCommand::CDashCommand(CKnowledge* const kb, const t_power pow,
			   const t_power max_pow,
			   const bool backp, const t_priority prio,
			   const CTime& s_time, const CTime& f_time)
  : CCommand(kb, CMD_Dash, prio, s_time, f_time),
    dash_type(DASH_Power),
    power(pow),
    max_power(max_pow),
    back_peddel(backp)
{
}

CDashCommand::CDashCommand(CKnowledge* const kb, const CPoint& dest,
			   const t_power max_pow,
			   const bool backp, const t_priority prio,
			   const CTime& s_time, const CTime& f_time)
  : CCommand(kb, CMD_Dash, prio, s_time, f_time),
    dash_type(DASH_Dest),
    max_power(max_pow),
    back_peddel(backp),
    destination(dest)
{
}

CDashCommand::CDashCommand(CKnowledge* const kb, const CVector& dir,
			   const t_power max_pow,
			   const bool backp, const t_priority prio,
			   const CTime& s_time, const CTime& f_time)
  : CCommand(kb, CMD_Dash, prio, s_time, f_time),
    dash_type(DASH_Dir),
    max_power(max_pow),
    back_peddel(backp),
    direction(dir)
{
}

CDashCommand::CDashCommand(CKnowledge* const kb, const float_type val,
			   const t_dash_type type,
			   const t_power max_pow, const bool backp, 
			   const t_priority prio, const CTime& s_time,
			   const CTime& f_time)
  : CCommand(kb, CMD_Dash, prio, s_time, f_time),
    dash_type(type),
    max_power(max_pow),
    back_peddel(backp)
{
  switch (type) {
    case DASH_Speed : speed = val; break;
    case DASH_Dist  : distance = val; break;
    default:
      dash_type = DASH_Power;
      power = val;      
  }
}

CDashCommand::CDashCommand(const CDashCommand& obj)
  : CCommand(obj),
    dash_type(obj.dash_type),
    power(obj.power),
    max_power(obj.max_power),
    back_peddel(obj.back_peddel),
    speed(obj.speed),
    destination(obj.destination),
    distance(obj.distance)
{
}

CDashCommand::~CDashCommand()
{
}

void CDashCommand::CopyCommand(const CDashCommand& obj)
{
  CCommand::CopyCommand(obj);
  dash_type = obj.dash_type;
  power = obj.power;
  max_power = obj.max_power;
  back_peddel = obj.back_peddel;
  speed = obj.speed;
  destination = obj.destination;
  distance = obj.distance;
}

CDashCommand& CDashCommand::operator =(const CDashCommand& obj)
{
  if ( this != &obj ) {
    CopyCommand(obj);
  }

  return *this;
}

void CDashCommand::DashPower(const t_power pow)
{
  dash_type = DASH_Power;
  power = pow;
}
  
void CDashCommand::DashSpeed(const float_type spd)
{
  dash_type = DASH_Speed;
  speed = spd;
}

void CDashCommand::DashDestination(const CPoint& dest)
{
  dash_type = DASH_Dest;
  destination = dest;
}

void CDashCommand::DashDistance(const float_type dist)
{
  dash_type = DASH_Dist;
  distance = dist;
}

void CDashCommand::UpdateAfterCommand(const int_type from,
				      const int_type to) const
{

  // First we need to normalize the power, since we can't be sure that the
  // power is actually within its bounds.
  float_type real_power = MinMax(k->GetParameters()->SP_minpower, power,
				 k->GetParameters()->SP_maxpower);
  float_type needed_power = real_power;
  const float_type max_needed_power = k->theAgent().GetStamina(from)
                                       - k->GetParameters()->SP_stamina_min;
  // If we do a negative dash the cost is doubled.
  if ( power < 0 ) {
    needed_power *= -2;
  }

  // Check if I have enough stamina to do the dash,
  // otherwise decrease the power so that the stamina
  // will be SP_stamina_min afterwards.
  if ( needed_power > max_needed_power ) {
    needed_power = max_needed_power;
    // Since the cost of negative dash is doubled half the
    // real power if negative dash.
    real_power = ( power < 0 ) ? needed_power/-2 : needed_power;
  }

  // A dash sets the agent's acceleration in the direction of the body.
  CVector acc(k->theAgent().GetBodyDir(from),
	      real_power * k->theAgent().GetEffort(from)
	                 * k->GetParameters()->SP_dash_power_rate);
  k->theAgent().SetAcceleration(acc, to);

  // The stamina is decreased by the needed_power.
  k->theAgent().SetStamina(k->theAgent().GetStamina(from) - needed_power, from);
  k->theAgent().UpdateMovement(from, to);
}

bool CDashCommand::Validate(const int_type hist)
{
  
  // First calculate the power I want to dash with
  switch(dash_type) {
  case DASH_Power : // Dash with power already set
    break;
  case DASH_Speed : // Dash with an intended speed
    DashSpeed(hist);
    break;
  case DASH_Dest : // Dash with an intended destination
    DashDestination(hist);
    break;
  case DASH_Dir : // Dash with an intended direction
    DashDirection(hist);
    break;
  case DASH_Dist : // Dash with an intended distance
    DashDistance(hist);
    break;
  default : 
    std::cerr << "Unknown dash_type in DashCommand.\n";
    return false;
  }

  // Then validate it.
  // Validate makes the power equal to the actual power needed to
  // do a dash with the same "real" power.

  // Since I can move before kick off, dash is a waste of
  // stamina (and uncertain).
  if ( k->theGame().GetPlayMode() == PM_Before_Kick_Off ) {
#ifdef DEBUG
    std::cerr << "No dash since game is stopped!" << std::endl;
#endif
    return false;
  }

  // Since power is multiplied with the effort, divide it with
  // the effort to acount for it.
  power /= k->theAgent().GetEffort(hist);

  // Make sure that the power is within it's range.
  power = MinMax(k->GetParameters()->SP_minpower, power,
		 k->GetParameters()->SP_maxpower);

  // Check if I have enough stamina to do the dash,
  // otherwise decrease the power so that the stamina
  // will be SP_stamina_min afterwards.
  float_type stamina = k->theAgent().GetStamina(hist);
  float_type needed_power = power;
  const float_type max_needed_power = stamina - k->GetParameters()->SP_stamina_min;

  // If we do a negative dash the cost is doubled.
  if ( power < 0 ) {
    needed_power *= -2;
  }

  // If this is wanted put it in CPrimAct, or??
#ifdef _CONSERVE_STAMINA_
  if (cmd_priority == PRIO_LOW) {
    if (k->theAgent().GetEffort(hist) < k->GetParameters()->SP_effort_max)
      needed_power = MinMax(0.0, needed_power, max_needed_power
			    - k->GetParameters()->SP_stamina_max
			    * k->GetParameters()->SP_effort_inc_thr);
    else
      needed_power = MinMax(0.0, needed_power, max_needed_power
			    - k->GetParameters()->SP_stamina_max
			    * k->GetParameters()->SP_effort_dec_thr);
  } else if (cmd_priority == PRIO_HIGH) {
    needed_power = Min(needed_power, max_needed_power);
    //} else if (GameAlmostOver()) {
    //needed_power = Min(needed_power, max_needed_power);
  } else {
    needed_power = MinMax(0.0, needed_power, max_needed_power
			  - k->GetParameters()->SP_stamina_max
			  * k->GetParameters()->SP_recover_dec_thr);    
  }
  if ( needed_power < 0 ) {
#ifdef DEBUG
    std::cout << "Needed_power negative!! power " << power
	      << " stamina " << stamina << " needed_power "
	      << needed_power << " max_needed_power "
	      << max_needed_power << " effort "
	      << k->theAgent().GetEffort(hist)
	      << " stamina_max " << k->GetParameters()->SP_stamina_max
	      << " recover_dec_thr " << k->GetParameters()->SP_recover_dec_thr
	      << std::endl;
#endif
    needed_power = 0;
  }
  power = ( power < 0 ) ? needed_power/-2 : needed_power;
#else
  if ( needed_power > max_needed_power ) {
    needed_power = max_needed_power;
    // Since the cost of negative dash is doubled half the
    // power if a negative dash.
    power = ( power < 0 ) ? needed_power/-2 : needed_power;
  }  
#endif
  
  /* If this is wanted, put it in CPrimAct.
     if (Abs(power) < k->GetParameters()->RS_min_dash_power)
       power = 0;
   */

  // To minimize waste of stamina, check that the dash will not
  // generate a speed that's greater than the max speed.
  // This is done by simulating a dash.
  CVector acc(k->theAgent().GetBodyDir(hist),
	      power * k->GetParameters()->SP_dash_power_rate);
  CVector v = k->GetParameters()->SP_player_decay
               * (k->theAgent().GetAbsoluteSpeedVector(hist) + acc);
  if (v.GetLength() > k->GetParameters()->SP_player_speed_max) {
#ifdef DEBUG
    std::cout << "Too much power: " << power;
#endif
    speed = k->GetParameters()->SP_player_speed_max;
    dash_type = DASH_Speed;
    Validate(hist);
#ifdef DEBUG
    std::cout << " new power: " << power << std::endl;
#endif
  }

  return true;
}

void CDashCommand::DashSpeed(const int_type hist)
{
  power = speed / k->GetParameters()->SP_player_decay - k->theAgent().GetSpeed(hist);
  power = power / k->GetParameters()->SP_dash_power_rate;

  if ( power < 0 && speed >= 0 )
    power = k->GetParameters()->SP_maxpower;
  else
    power = Min(power, max_power);

  if ( back_peddel )
    power = -power;

#ifdef DEBUG
  std::cout << "Dash power = " << power << std::endl;
#endif

}

void CDashCommand::DashDestination(const int_type hist)
{
  speed = k->theAgent().GetDistanceTo(destination, hist);
  DashSpeed(hist);

  /*
  if ( Abs(k->theAgent().GetRelativeDirectionTo(destination, hist)) > 90 )
    power = -power;
  */
}

void CDashCommand::DashDirection(const int_type hist)
{
  speed = direction.GetLength();
  DashSpeed(hist);

  /*
  if ( Abs(direction.GetAngle()) > 90 )
    power = -power;
  */
}

void CDashCommand::DashDistance(const int_type hist)
{
  // Since speed is the distance travel in one turn distance and speed
  // are equivalent when for only one turn.
  speed = distance;
  DashSpeed(hist);
}

const std::string CDashCommand::ServerString() const
{
  return "(dash " + ToString(power) + ")";
}

