#ifndef _CGAMEOBJECT_H_
#define _CGAMEOBJECT_H_

#include "soc_types.h"
#include "CConfidence.h"
#include "CAngleDeg.h"
#include "CPoint.h"
#include "CVector.h"
#include "CTime.h"
#include <vector>
#include <iostream>
//#include "CRectangle.h"

static const int_type DEFAULT_HIST_LENGTH = 2;
static const int_type PREV = -1;
static const int_type NOW  =  0;
static const int_type NEXT =  1;
static const int_type CLEAN_EST = 2;
static const int_type CURRENT_EST = 3;
static const int_type NEXT_EST = 4;

// This class might throw out_of_range exceptions, although I don't
// think there should ever be such a problem since everything goes through
// GetIndex which never goes out of bounds.

class CGameObject
{
public:
  CGameObject(const int_type hist_length = DEFAULT_HIST_LENGTH);
  CGameObject(const CGameObject& obj);
  //  ~CGameObject();

  CGameObject& operator= (const CGameObject& obj);

  void Copy(const int_type from, const int_type to);
  void CopyFrom(const CGameObject& obj, const int_type from, const int_type to);
  
  void ResetAll();
  void Reset();
  void Reset(const int_type i);

  void NewCycle();

  void UpdateVector(const CPoint& agent_pos, const CAngleDeg& agent_body_dir,
		    const int_type = NEXT);
  void UpdateConfidence(const int_type from = NOW, const int_type to = NEXT);

  inline void SetConfidenceDecay(const float_type conf_dec);
  inline float_type GetConfidenceDecay() const;

  // hist = number of turns ahead.
  // hist = -1 means the last cycles info (= PREV).
  // hist =  0 means the current cycles info (= NOW).
  // hist =  1 means the next cycles info (= NEXT).
  inline float_type GetPosX(const int_type hist = NOW) const;
  inline float_type GetPosY(const int_type hist = NOW) const;
  inline void SetPosition(const CPoint& pos, const int_type hist = NEXT);
  inline CPoint GetPosition(const int_type hist = NOW) const;
  inline void CopyPosition(const int_type from, const int_type to);

  inline void SetDirection(const CAngleDeg& dir, const int_type hist = NEXT);
  inline CAngleDeg GetDirection(const int_type hist = NOW) const;
  inline void SetDistance(const float_type dist, const int_type hist = NEXT);
  inline float_type GetDistance(const int_type hist = NOW) const;
  inline void SetVector(const CVector& v, const int_type hist = NEXT);
  inline CVector GetVector(const int_type hist = NOW) const;
  inline void CopyVector(const int_type from, const int_type to);

  inline void SetObsTime(const CTime& time, const int_type hist = NEXT);
  inline CTime GetObsTime(const int_type hist = NOW) const;
  inline void CopyObsTime(const int_type from, const int_type to);
  
  inline void SetConfidence(const CConfidence& new_confidence,
			    const int_type hist = NEXT);
  inline CConfidence GetConfidence(const int_type hist = NOW) const;
  inline void CopyConfidence(const int_type from, const int_type to);

  CVector GetAbsoluteVectorTo(const CPoint& pt, const int_type hist = NOW) const;
  CAngleDeg GetAbsoluteDirectionTo(const CPoint& pt, const int_type hist = NOW) const;
  float_type GetDistanceTo(const CPoint& pt, const int_type hist = NOW) const;
  bool ShouldBeSeen(const CAngleDeg& face_dir, const CAngleDeg& view_width,
		    const int_type hist = NOW) const;
  // InSector(const CSector& sec, const int_type hist = NOW) const;
  // inline InRectangle(const CRectangle& rect, const int_type hist = NOW) const
  //  { return rect.Inside(positions[GetIndex(hist)]); };

  //bool InsideField(const int_type hist) const;
  
  void Print(const int_type hist = UNKNOWN_INT_TYPE) const;

protected:
  int_type max_hist_length;
  int_type max_length;
  int_type current_obs;

  std::vector<CPoint> positions;
  std::vector<CVector> vectors;
  std::vector<CTime> obs_times;
  std::vector<CConfidence> confidences;
  float_type confidence_decay;
  
  static const CPoint DEFAULT_POSITION;
  static const CVector DEFAULT_VECTOR;
  static const CTime DEFAULT_OBS_TIME;
  //  static const CConfidence DEFAULT_CONFIDENCE;
  static const float_type DEFAULT_CONFIDENCE_DECAY;
  
  int_type GetIndex(const int_type i = NOW) const;

  void CopyData(const CGameObject&);
  
};

inline int_type CGameObject::GetIndex(const int_type i) const
{
  return (current_obs + i + max_length) % max_length;
}

inline void CGameObject::UpdateConfidence(const int_type from, const int_type to)
{
  confidences[GetIndex(to)] = confidences[GetIndex(from)] * confidence_decay;
} 

inline void CGameObject::SetConfidenceDecay(const float_type conf_dec)
{
  confidence_decay = conf_dec;
}

inline float_type CGameObject::GetConfidenceDecay() const
{
  return confidence_decay;
}

inline float_type CGameObject::GetPosX(const int_type hist) const
{
  return positions[GetIndex(hist)].GetX();
}

inline float_type CGameObject::GetPosY(const int_type hist) const
{
  return positions[GetIndex(hist)].GetY();
}

inline void CGameObject::SetPosition(const CPoint& pos, const int_type hist)
{
  positions[GetIndex(hist)] = pos;
}

inline CPoint CGameObject::GetPosition(const int_type hist) const
{
  return positions[GetIndex(hist)];
}

inline void CGameObject::CopyPosition(const int_type from, const int_type to)
{
  positions[GetIndex(to)] = positions[GetIndex(from)];
}

inline void CGameObject::SetDirection(const CAngleDeg& dir, const int_type hist)
{
  vectors[GetIndex(hist)].SetAngle(dir);
}

inline CAngleDeg CGameObject::GetDirection(const int_type hist) const
{
  return vectors[GetIndex(hist)].GetAngle();
}

inline void CGameObject::SetDistance(const float_type dist, const int_type hist)
{
  vectors[GetIndex(hist)].SetLength(dist);
}

inline float_type CGameObject::GetDistance(const int_type hist) const
{
  return vectors[GetIndex(hist)].GetLength();
}

inline void CGameObject::SetVector(const CVector& v, const int_type hist)
{
  vectors[GetIndex(hist)] = v;
}

inline CVector CGameObject::GetVector(const int_type hist) const
{
  return vectors[GetIndex(hist)];
}

inline void CGameObject::CopyVector(const int_type from, const int_type to)
{
  vectors[GetIndex(to)] = vectors[GetIndex(from)];
}

inline void CGameObject::SetObsTime(const CTime& time, const int_type hist)
{
  obs_times[GetIndex(hist)] = time;
}

inline CTime CGameObject::GetObsTime(const int_type hist) const
{
  return obs_times[GetIndex(hist)];
}
  
inline void CGameObject::CopyObsTime(const int_type from, const int_type to)
{
  obs_times[GetIndex(to)] = obs_times[GetIndex(from)];
}

inline void CGameObject::SetConfidence(const CConfidence& new_confidence,
				       const int_type hist)
{
  confidences[GetIndex(hist)] = new_confidence;
}

inline CConfidence CGameObject::GetConfidence(const int_type hist) const
{
  return confidences[GetIndex(hist)];
}

inline void CGameObject::CopyConfidence(const int_type from, const int_type to)
{
  confidences[GetIndex(to)] = confidences[GetIndex(from)];
}

inline bool CGameObject::ShouldBeSeen(const CAngleDeg& face_dir,
				      const CAngleDeg& view_width,
				      const int_type hist) const
{
  return ( Abs(GetDirection(hist) - face_dir) < view_width/2 );
}

/*
inline bool CGameObject::SetInsideField(const int_type hist)
{
  inside_field = new_val;
}

inline bool CGameObject::InsideField(const int_type hist) const
{
  return inside_field;
}
*/

#endif
