#include "CPlayerObject.h"
#include "compiler_directives.h"

#ifdef _REQUIRE_PARAMETERS_IN_OBJECTS_
#  include "soc_assert.h"
#endif

const t_team CPlayerObject::DEFAULT_TEAM = T_Unknown;
const t_side CPlayerObject::DEFAULT_SIDE = S_Unknown;
const t_unum CPlayerObject::DEFAULT_U_NUMBER = UNKNOWN_UNUM;
const CAngleDeg CPlayerObject::DEFAULT_BODY_DIR = UNKNOWN_DEG;
const CAngleDeg CPlayerObject::DEFAULT_NECK_DIR = UNKNOWN_DEG;
const CAngleDeg CPlayerObject::DEFAULT_FACE_DIR = UNKNOWN_DEG;

CPlayerObject::CPlayerObject(const CParameters* const p, const int_type hist_length,
			     const t_team new_team, const t_side new_side,
			     const t_unum new_u_number)
  : CMobileObject(hist_length),
    params(p),
    have_parameters(false),
    team(new_team),
    side(new_side),
    u_number(new_u_number),
    body_dirs(max_length),
    neck_dirs(max_length),
    face_dirs(max_length)
{
#ifdef _REQUIRE_PARAMETERS_IN_OBJECTS_
  Assert(params != NULL,
	 "Must include a pointer to a CParameter object in CPlayerObject!");
  have_parameters = true;
#else
  if (params == NULL) {
    have_parameters = false;
  } else {
    have_parameters = true;
  }
#endif  

#ifndef _REQUIRE_PARAMETERS_IN_OBJECTS_
  if ( have_parameters ) {
#endif
    SetSpeedMax(params->SP_player_speed_max);
    SetSpeedDecay(params->SP_player_decay); 
#ifndef _REQUIRE_PARAMETERS_IN_OBJECTS_
  }
#endif
  
  for (int_type i=0; i < max_length; ++i) {
    Reset(i);
  }
  
}  

CPlayerObject::CPlayerObject(const CParameters* const p, const t_team new_team,
			     const t_side new_side, const t_unum new_u_number,
			     const int_type hist_length)
  : CMobileObject(hist_length),
    params(p),
    have_parameters(false),
    team(new_team),
    side(new_side),
    u_number(new_u_number),
    body_dirs(max_length),
    neck_dirs(max_length),
    face_dirs(max_length)
{
#ifdef _REQUIRE_PARAMETERS_IN_OBJECTS_
  Assert(params != NULL,
	 "Must include a pointer to a CParameter object in CPlayerObject!");
  have_parameters = true;
#else
  if (params == NULL) {
    have_parameters = false;
  } else {
    have_parameters = true;
  }
#endif  

#ifndef _REQUIRE_PARAMETERS_IN_OBJECTS_
  if ( have_parameters ) {
#endif
    SetSpeedMax(params->SP_player_speed_max);
    SetSpeedDecay(params->SP_player_decay); 
#ifndef _REQUIRE_PARAMETERS_IN_OBJECTS_
  }
#endif
  
  for (int_type i=0; i < max_length; ++i) {
    Reset(i);
  }
  
}

CPlayerObject::CPlayerObject(const CPlayerObject& obj)
  : CMobileObject(obj),
    params(obj.params),
    have_parameters(obj.have_parameters),
    team(obj.team),
    side(obj.side),
    u_number(obj.u_number),
    body_dirs(obj.body_dirs),
    neck_dirs(obj.neck_dirs),
    face_dirs(obj.face_dirs)
{

}  

/*
CPlayerObject::~CPlayerObject()
{

}
*/

void CPlayerObject::CopyData(const CPlayerObject& obj)
{
  CMobileObject::CopyData(obj);
  params          = obj.params;
  have_parameters = obj.have_parameters;
  team      = obj.team;
  side      = obj.side;
  u_number  = obj.u_number;
  body_dirs = obj.body_dirs;
  neck_dirs = obj.neck_dirs;
  face_dirs = obj.face_dirs;      
}

CPlayerObject& CPlayerObject::operator= (const CPlayerObject& obj)
{

  if (this != &obj) {
    CopyData(obj);
  }

  return *this;
  
}

void CPlayerObject::Copy(const int_type from, const int_type to)
{
  CMobileObject::Copy(from, to);
  CopyBodyDir(from, to);
  CopyNeckDir(from, to);
  CopyFaceDir(from, to);
}

void CPlayerObject::CopyFrom(const CPlayerObject& plr, const int_type from,
			     const int_type to)
{
  CMobileObject::CopyFrom(plr, from, to);
  body_dirs[GetIndex(to)] = plr.body_dirs[GetIndex(from)];
  neck_dirs[GetIndex(to)] = plr.neck_dirs[GetIndex(from)];
  face_dirs[GetIndex(to)] = plr.face_dirs[GetIndex(from)];
}

void CPlayerObject::ResetAll()
{
  Reset();
  
  for (int_type i = 0; i < max_length; ++i)
    Reset(i);

}

void CPlayerObject::Reset()
{
  CMobileObject::Reset();

  team = DEFAULT_TEAM;
  side = DEFAULT_SIDE;
  u_number = DEFAULT_U_NUMBER;
}

void CPlayerObject::Reset(const int_type i)
{

  CMobileObject::Reset(i);
  body_dirs[GetIndex(i)] = DEFAULT_BODY_DIR;
  neck_dirs[GetIndex(i)] = DEFAULT_NECK_DIR;
  face_dirs[GetIndex(i)] = DEFAULT_FACE_DIR;
  
}

void CPlayerObject::NewCycle()
{
  CMobileObject::NewCycle();
  Copy(PREV, NOW);
}

// Redefined since we now have access to the player_speed_max parameter.
void CPlayerObject::SetAbsoluteSpeedVector(const CVector& v, const int_type hist)
{
  SetSpeed(v.GetLength(), hist),
  SetAbsoluteSpeedDirection(v.GetAngle(), hist);
}

// Redefined since we now have access to the player_speed_max parameter.
void CPlayerObject::SetSpeed(const t_speed& speed, const int_type hist)
{
#ifndef _REQUIRE_PARAMETERS_IN_OBJECTS_
  if ( have_parameters ) {
#endif
    CMobileObject::SetSpeed(MinMax(0.0, speed, params->SP_player_speed_max), hist);
#ifndef _REQUIRE_PARAMETERS_IN_OBJECTS_
  } else {
    CMobileObject::SetSpeed(speed, hist);
  }
#endif    
}

void CPlayerObject::SetTeam(const t_side our_side)
{
  if ( our_side == side )
    team = T_Our_Team;
  else if ( our_side == S_Unknown )
    team = T_Unknown;
  else
    team = T_Their_Team;
}

CVector CPlayerObject::GetRelativeVectorTo(const CPoint& pt, const int_type hist) const
{

  CVector vec = CGameObject::GetAbsoluteVectorTo(pt, hist);
  
  if ( body_dirs[GetIndex(hist)] == UNKNOWN_DEG ||
       vec == UNKNOWN_VECTOR )
    return UNKNOWN_VECTOR;
  else 
    return vec.Rotate(-body_dirs[GetIndex(hist)]);

}

CAngleDeg CPlayerObject::GetRelativeDirectionTo(const CPoint& pt, const int_type hist) const
{

  CAngleDeg ang = CGameObject::GetAbsoluteDirectionTo(pt, hist);

  if ( body_dirs[GetIndex(hist)] == UNKNOWN_DEG ||
       ang == UNKNOWN_DEG )
    return UNKNOWN_DEG;
  else
    return (ang - body_dirs[GetIndex(hist)]);

}

void CPlayerObject::Print(const int_type hist) const
{

  if (hist == UNKNOWN_INT_TYPE) {
    for (int_type i = -max_hist_length; i < max_length - max_hist_length; ++i) {
      CMobileObject::Print(i);  
      std::cout << "[" << i << "]"
		<< " BodyDir: " << GetBodyDir(i)
		<< " NeckDir: " << GetNeckDir(i)
		<< " FaceDir: " << GetFaceDir(i)
		<< std::endl;
    }
  } else {
    CMobileObject::Print(hist);
    std::cout << "[" << hist << "]"
	      << " BodyDir: " << GetBodyDir(hist)
	      << " NeckDir: " << GetNeckDir(hist)
	      << " FaceDir: " << GetFaceDir(hist)
	      << std::endl;
  }
  
}

